﻿using System;
using System.Threading;
using ServoControl;

namespace Control
{
    using System.Collections.Generic;
    using System.Diagnostics;
    using System.Linq;
    using System.Reflection;

    public class Automation
    {
        private readonly Control _control;
        private readonly Sensor _sensor;
        private readonly InverseKinematics _ik;

        public Automation(Control control, Sensor sensor, InverseKinematics ik)
        {
            _control = control;
            _sensor = sensor;
            _ik = ik;

            _iKvalues = new List<IkValue>();
        }


        private readonly Random _random = new Random();
        
        private bool _running;
        private List<IkValue> _iKvalues;

        private double _lastZ = 0;

        public void Start()
        {
            _running = true;

            var t = new Thread(() =>
            {
                _ik.TravelLengthX = 0;
                _ik.TravelLengthZ = 0;
                _ik.TravelRotationY = 0;

                _ik.GaitType = 1;
                _ik.LegLiftHeight = 15;
                _ik.BodyPosY = 85;
                _ik.NomGaitSpeed = 70;

                var ikValue = new IkValue();

                CenterYaw();

                //Main loop ------------------
                while (_running)
                {
                    var avoidDelay = _random.Next(5, 12);

                    _iKvalues.Add(ikValue);

                    if (_sensor.Roll > .2)
                    {
                        _ik.BodyRotZ1 = -(Map(_sensor.Roll, 20, 0, 1, 6));
                    }
                    else if (_sensor.Roll < -.2)
                    {
                        _ik.BodyRotZ1 = (Map(_sensor.Roll, 0, 20, 1, 6));
                    }
                    else
                        _ik.BodyRotZ1 = 0;

                    if (_sensor.Yaw > .2)
                    {
                        _ik.BodyRotX1 = -(Map(_sensor.Yaw, 20, 0, 1, 6));
                    }
                    else if (_sensor.Yaw < -.2)
                    {
                        _ik.BodyRotX1 = (Map(_sensor.Yaw, 0, 20, 1, 6));
                    }
                    else
                        _ik.BodyRotX1 = 0;

                    if (_iKvalues.Count > 5)
                    {
                        _ik.TravelLengthX = _iKvalues.Last().X;
                        _ik.TravelLengthZ = _iKvalues.Last().Z;

                        _lastZ = _iKvalues.Last().Z;

                        if (_iKvalues.Last().Z > 0)
                        {
                            while (!_sensor.WarningRear && !_sensor.WarningRearLeft && !_sensor.WarningRearRight)
                            {
                                Thread.Sleep(1);
                            }

                            if (!_sensor.WarningRearLeft)
                                _ik.TravelRotationY = -2;
                            else if (!_sensor.WarningRearRight)
                                _ik.TravelRotationY = 2;

                            Thread.Sleep(100 * avoidDelay);
                        }
                        else
                        {
                            if (_iKvalues.Last().Heading != Direction.Center)
                                Turn(_iKvalues.Last().Heading);
                            else
                            {
                                _ik.TravelRotationY = 0;
                            }
                        }
                    }

                    ikValue = new IkValue();

                    Thread.Sleep(40);

                    if (!_sensor.WarningCenter && !_sensor.WarningLeft && !_sensor.WarningRight)
                    {
                        Debug.WriteLine("Forward");
                        ikValue.Heading = Direction.Center;
                        ikValue.Z = -(Map(_sensor.DistanceSensorCenter, 10, 120, 15, 70) + 20);
                        _ik.BodyRotZ1 = 0;
                    }
                    else if (_sensor.WarningCenter || (_sensor.WarningLeft && _sensor.WarningRight))
                    {
                        ikValue.Z = Map(_sensor.DistanceSensorCenter, 10, 312, 15, 90) + 20;
                        _ik.BodyRotZ1 = 0;
                    }
                    else if (_sensor.WarningLeft && !_sensor.WarningRight)
                    {
                        Debug.WriteLine("Turn right");

                        ikValue.Heading = Direction.Right;
                    }
                    else if (_sensor.WarningRight && !_sensor.WarningLeft)
                    {
                        Debug.WriteLine("Turn left");

                        ikValue.Heading = Direction.Left;
                    }

                    if (_sensor.WarningFarLeft)
                    {
                        ikValue.X = Map(_sensor.DistanceSensorCenter, 5, 100, 15, 60);
                        //_ik.BodyRotZ1 = 5;
                    }
                    else
                        _ik.BodyRotZ1 = 0;

                    if (_sensor.WarningFarRight)
                    {
                        ikValue.X = -Map(_sensor.DistanceSensorCenter, 5, 100, 15, 60);
                        //_ik.BodyRotZ1 = -5;
                    }
                    else
                        _ik.BodyRotZ1 = 0;

                }
            }) { IsBackground = true };
            t.Start();
        }

        //public void StartNew()
        //{
        //    var n = MapNegative(200, 0, 300, -90, 90);
        //    var n1 = MapNegative(99, 0, 300, -90, 90);
        //    var n2 = MapNegative(1, 0, 300, -90, 90);
        //    var n3 = MapNegative(288, 0, 300, -90, 90);

        //    _running = true;

        //    var r = CenterYaw();

        //    var t = new Thread(() =>
        //    {
        //        _ik.TravelLengthX = 0;
        //        _ik.TravelLengthZ = 0;
        //        _ik.TravelRotationY = 0;

        //        _ik.GaitType = 1;
        //        _ik.LegLiftHeight = 15;
        //        _ik.BodyPosY = 110;
        //        _ik.NomGaitSpeed = 70;

        //        var ikValue = new IkValue();

        //        //Main loop ------------------
        //        while (_running)
        //        {
        //            var avoidDelay = _random.Next(5, 12);

        //            _iKvalues.Add(ikValue);

        //            if (_iKvalues.Count > 5)
        //            {
        //                _ik.TravelLengthX = _iKvalues.Last().X;
        //                _ik.TravelLengthZ = _iKvalues.Last().Z;

        //                _lastZ = _iKvalues.Last().Z;

        //                if (_iKvalues.Last().Z > 0)
        //                {
        //                    //var currentRearDistance = _sensor.DistanceSensorRear;

        //                    //while (_sensor.DistanceSensorRear >= (currentRearDistance - 10))
        //                    //{
        //                    //    Thread.Sleep(1);
        //                    //}

        //                    Thread.Sleep(350 * avoidDelay);

        //                    if (!_sensor.WarningRearLeft)
        //                        _ik.TravelRotationY = -2;
        //                    else if (!_sensor.WarningRearRight)
        //                        _ik.TravelRotationY = 2;

        //                    Thread.Sleep(100 * avoidDelay);
        //                }
        //                else
        //                {
        //                    if (_iKvalues.Last().Heading != Direction.Center)
        //                        Turn(_iKvalues.Last().Heading);
        //                    else
        //                    {
        //                        _ik.TravelRotationY = 0;
        //                    }
        //                }
        //            }

        //            ikValue = new IkValue();

        //            Thread.Sleep(20);

        //            if (!_sensor.WarningCenter && !_sensor.WarningLeft && !_sensor.WarningRight)
        //            {
        //                Debug.WriteLine("Forward");
        //                ikValue.Heading = Direction.Center;
        //                ikValue.Z = -(Map(_sensor.DistanceSensorCenter, 10, 120, 15, 70) + 20);
        //                _ik.BodyRotZ1 = 0;
        //            }
        //            else if (_sensor.WarningCenter || (_sensor.WarningLeft && _sensor.WarningRight))
        //            {
        //                ikValue.Z = Map(_sensor.DistanceSensorCenter, 10, 312, 15, 90) + 20;
        //                _ik.BodyRotZ1 = 0;
        //            }
        //            else if (_sensor.WarningLeft && !_sensor.WarningRight)
        //            {
        //                Debug.WriteLine("Turn right");

        //                ikValue.Heading = Direction.Right;
        //            }
        //            else if (_sensor.WarningRight && !_sensor.WarningLeft)
        //            {
        //                Debug.WriteLine("Turn left");

        //                ikValue.Heading = Direction.Left;
        //            }

        //            if (_sensor.WarningFarLeft)
        //            {
        //                ikValue.X = Map(_sensor.DistanceSensorCenter, 10, 312, 15, 60);
        //                //_ik.BodyRotZ1 = 10;
        //            }

        //            if (_sensor.WarningFarRight)
        //            {
        //                ikValue.X = -Map(_sensor.DistanceSensorCenter, 10, 312, 15, 60);
        //                //_ik.BodyRotZ1 = -10;
        //            }

        //        }
        //    }) { IsBackground = true };
        //    t.Start();
        //}

        bool CenterYaw()
        {
            Debug.WriteLine("Finding center yaw");

            while (_sensor.Yaw > 5 || _sensor.Yaw < -5)
            {
                if (_sensor.Yaw > 5)
                {
                    _ik.TravelLengthZ = 2;
                    _ik.TravelLengthX = 0;
                    _ik.TravelRotationY = -1;
                }
                else if (_sensor.Yaw < -5)
                {
                    _ik.TravelLengthZ = 2;
                    _ik.TravelLengthX = 0;
                    _ik.TravelRotationY = 1;
                }
                else
                {
                    _ik.TravelLengthZ = 0;
                    _ik.TravelLengthX = 0;
                    _ik.TravelRotationY = 0;
                    break;
                }
            }

            Debug.WriteLine("Found center yaw " + _sensor.Yaw);

            return true;
        }

        Direction FindOpening()
        {
            if (!_sensor.WarningCenter)
                return Direction.Center;

            if (!_sensor.WarningLeft)
                return Direction.LeftMid;

            if (!_sensor.WarningFarLeft)
                return Direction.Left;

            if (!_sensor.WarningRight)
                return Direction.RightMid;

            if (!_sensor.WarningFarRight)
                return Direction.Right;

            if (!_sensor.WarningRearLeft)
                return Direction.ReverseLeftMid;

            if (!_sensor.WarningRearRight)
                return Direction.ReverseRightMid;

            if (!_sensor.WarningRear)
                return Direction.Reverse;

            return Direction.Stop;
        }

        readonly object _turnLock = new object();

        bool Turn(Direction direction)
        {
            lock (_turnLock)
            {
                if (direction == Direction.Left)
                {
                    _ik.TravelLengthZ = _lastZ;
                    _ik.TravelLengthX = -10;
                    _ik.TravelRotationY = -1;
                }
                if (direction == Direction.Right)
                {
                    _ik.TravelLengthZ = _lastZ;
                    _ik.TravelLengthX = 10;
                    _ik.TravelRotationY = 1;
                }
          
                Thread.Sleep(100);
            }
            return true;
        }

        public void Stop()
        {
            _running = false;

            Thread.Sleep(100);
            _ik.TravelRotationY = 0;
            _ik.TravelLengthZ = 0;
            _ik.TravelLengthX = 0;
        }

        public static double Map(double x, double inMin, double inMax, double outMin, double outMax)
        {
            var r = (x - inMin) * (outMax - outMin) / (inMax - inMin) + outMin;
            return Math.Abs(r);
        }

        public static double MapNegative(double x, double inMin, double inMax, double outMin, double outMax)
        {
            var r = (x - inMin) * (outMax - outMin) / (inMax - inMin) + outMin;
            return r;
        }
    }

    internal class IkValue
    {
        public IkValue()
        {
            X = 0;
            Z = 0;
            Heading = Direction.Center;
        }

        public double X { get; set; }
        public double Z { get; set; }
        public Direction Heading { get; set; }
    }
}