﻿#include "I2Cdev.h"
#include "MPU6050_6Axis_MotionApps20.h"
#include <HMC5883L_Simple.h>
#include <Wire.h>

MPU6050 mpu;

#define LED_PIN 13
bool blinkState = false;
bool dmpReady = false;  // set true if DMP init was successful

uint8_t mpuIntStatus;   // holds actual interrupt status byte from MPU
uint8_t devStatus;      // return status after each device operation (0 = success, !0 = error)
uint16_t packetSize;    // expected DMP packet size (default is 42 bytes)
uint16_t fifoCount;     // count of all bytes currently in FIFO
uint8_t fifoBuffer[64]; // FIFO storage buffer

// orientation/motion vars
Quaternion q;           // [w, x, y, z]         quaternion container
VectorInt16 aa;         // [x, y, z]            accel sensor measurements
VectorInt16 aaReal;     // [x, y, z]            gravity-free accel sensor measurements
VectorInt16 aaWorld;    // [x, y, z]            world-frame accel sensor measurements
VectorFloat gravity;    // [x, y, z]            gravity vector

float euler[3];         // [psi, theta, phi]    Euler angle container
float ypr[3];           // [yaw, pitch, roll]   yaw/pitch/roll container and gravity vector

float Yaw;
float Pitch;
float Roll;
float Heading;

int Left = 0;
int Right = 0;
int Center = 0;

int FarLeft = 0;
int FarRight = 0;

int RearLeft = 0;
int RearRight = 0;
int Rear = 0;

volatile bool mpuInterrupt = false;     // indicates whether MPU interrupt pin has gone high

HMC5883L_Simple Compass;

void setup()
{
  pinMode(28, INPUT); 	//sonar echo Center
  pinMode(26, OUTPUT);	//sonar trig

  pinMode(22, INPUT); 	//sonar echo brwn Right Front
  pinMode(24, OUTPUT);	//sonar trig red

  pinMode(37, INPUT); 	//sonar echo purple Left Front
  pinMode(35, OUTPUT);	//sonar trig blue

  pinMode(33, INPUT); 	//sonar echo gray Left
  pinMode(31, OUTPUT);	//sonar trig white

  pinMode(42, INPUT); 	//sonar echo Right
  pinMode(40, OUTPUT);	//sonar trig

  pinMode(38, INPUT); 	//sonar echo Right Rear
  pinMode(36, OUTPUT);	//sonar trig

  pinMode(51, INPUT); 	//sonar echo Left Rear
  pinMode(49, OUTPUT);	//sonar trig

  pinMode(45, INPUT); 	//sonar echo Rear
  pinMode(47, OUTPUT);	//sonar trig

  Wire.begin();
  TWBR = 24; // 400kHz I2C clock (200kHz if CPU is 8MHz)

  Serial.begin(38400); //Debug
  Serial1.begin(38400); //PC data

  Compass.SetDeclination(8, 36, 'E');
  Compass.SetSamplingMode(COMPASS_SINGLE);
  Compass.SetScale(COMPASS_SCALE_130);
  Compass.SetOrientation(COMPASS_HORIZONTAL_X_NORTH);


  mpu.initialize();
  devStatus = mpu.dmpInitialize();

  // supply your own gyro offsets here, scaled for min sensitivity
  mpu.setXGyroOffset(220);
  mpu.setYGyroOffset(76);
  mpu.setZGyroOffset(-85);
  mpu.setZAccelOffset(1788); // 1688 factory default for my test chip
  mpu.setDMPEnabled(true);
  attachInterrupt(0, dmpDataReady, RISING);
  mpuIntStatus = mpu.getIntStatus();
  packetSize = mpu.dmpGetFIFOPacketSize();

  pinMode(LED_PIN, OUTPUT);
}

void loop()
{
  Center = Ping(26, 28, 74);
  Update();
 
  FarLeft = Ping(31, 33, 74);
  Update();

  Right = Ping(24, 22, 74);
  Update();
 
  Left = Ping(35, 37, 74);   
  Update();
   
  RearLeft = Ping(49, 51, 74);
  Update();
  
  FarRight = Ping(40, 42, 74);
  Update();
  
  RearRight = Ping(36, 38, 74);
  Update();
 
  Rear = Ping(47, 45, 74);
  Update();

  blinkState = !blinkState;
  digitalWrite(LED_PIN, blinkState);
}

void UpdateYPRH()
{
  mpuInterrupt = false;

  mpuIntStatus = mpu.getIntStatus();

  fifoCount = mpu.getFIFOCount();

  while (fifoCount < packetSize) fifoCount = mpu.getFIFOCount();

  mpu.getFIFOBytes(fifoBuffer, packetSize);

  if ((mpuIntStatus & 0x10) || fifoCount == 1024)
  {
    mpu.resetFIFO();
  }
  else if (mpuIntStatus & 0x02)
  {
    mpu.dmpGetQuaternion(&q, fifoBuffer);
    mpu.dmpGetGravity(&gravity, &q);
    mpu.dmpGetYawPitchRoll(ypr, &q, &gravity);

    Yaw = ypr[0] * 180 / M_PI;
    Pitch = ypr[2] * 180 / M_PI;
    Roll = ypr[1] * 180 / M_PI;
  }
  
   Heading = Compass.GetHeadingDegrees();
}

void dmpDataReady()
{
  mpuInterrupt = true;
}

int Ping(int trigPin, int echoPin, int div)
{
  digitalWrite(trigPin, LOW);                           		// Set the trigger pin to low for 2uS
  delayMicroseconds(2);
  digitalWrite(trigPin, HIGH);                          		// Send a 10uS high to trigger ranging
  delayMicroseconds(10);
  digitalWrite(trigPin, LOW);                     		// Send pin low again
  return (pulseIn(echoPin, HIGH, 32000) / div / 2);
}

void Update()
{
  Serial.println(String(Center) + ":" + String(Right) + ":" + String(Left) + ":" + String(FarRight) + ":" + String(FarLeft) + ":" + String(RearLeft) + ":" + String(RearRight) + ":" + String(Rear) + ":" + String(Yaw) + ":" + String(Pitch) + ":" + String(Roll) + ":" + String(Heading) );
  Serial1.println(String(Center) + ":" + String(Right) + ":" + String(Left) + ":" + String(FarRight) + ":" + String(FarLeft) + ":" + String(RearLeft) + ":" + String(RearRight) + ":" + String(Rear) + ":" + String(Yaw) + ":" + String(Pitch) + ":" + String(Roll) + ":" + String(Heading) );

  if (mpuInterrupt) UpdateYPRH();

  delay(40);
}