﻿using System;
using System.Threading;
using ServoControl;

namespace Control
{
    //C# Port by Larry W of code originally written by Jeroen Janssen [aka Xan] and Kurt Eckhardt(KurtE)
    //This is a work in progress!
    //walkerDefaults.config needs to be copied to the running directory
    //Format of file is, min,max,servo,x,x| x can be set to 0 as they are no longer used. Really the important values are the servo numbers
    //1191,1791,7,0,2000|1318,2068,6,0,2000|1574,1974,5,0,2000
    public class InverseKinematics
    {
        private readonly Control _control;
        private readonly Sensor _sensor;

        public InverseKinematics(Control control, Sensor sensor)
        {
            _control = control;
            _sensor = sensor;
        }

        private const short CRrFemurHornOffset1 = 0;
        private const short CRmFemurHornOffset1 = 0;
        private const short CRfFemurHornOffset1 = 0;
        private const short CLrFemurHornOffset1 = 0;
        private const short CLmFemurHornOffset1 = 0;
        private const short CLfFemurHornOffset1 = 0;
        private const int CPfConst = 592; //old 650 ; 900*(1000/cPwmDiv)+cPFConst must always be 1500
        private const int CPwmDiv = 991; //old 1059;
        private const double CTravelDeadZone = 1;
        private const double C1Dec = 10;
        private const double C2Dec = 100;
        private const double C4Dec = 10000;
        private const double C6Dec = 1000000;
        private const int CLf = 5;
        private const int CLm = 4;
        private const int CLr = 3;
        private const int CRf = 2;
        private const int CRm = 1;
        private const int CRr = 0;
        private const int CoxaMin = -700; //-650
        private const int CoxaMax = 700; //650
        private const int FemurMin = -1050; //-1050
        private const int FemurMax = 350; //150
        private const int TibiaMin = -450; //-450
        private const int TibiaMax = 350; //350
        private const int CRrCoxaMin1 = CoxaMin; //Mechanical limits of the Right Rear Leg
        private const int CRrCoxaMax1 = CoxaMax;
        private const int CRrFemurMin1 = FemurMin;
        private const int CRrFemurMax1 = FemurMax;
        private const int CRrTibiaMin1 = TibiaMin;
        private const int CRrTibiaMax1 = TibiaMax;
        private const int CRmCoxaMin1 = CoxaMin; //Mechanical limits of the Right Middle Leg
        private const int CRmCoxaMax1 = CoxaMax;
        private const int CRmFemurMin1 = FemurMin;
        private const int CRmFemurMax1 = FemurMax;
        private const int CRmTibiaMin1 = TibiaMin;
        private const int CRmTibiaMax1 = TibiaMax;
        private const int CRfCoxaMin1 = CoxaMin; //Mechanical limits of the Right Front Leg
        private const int CRfCoxaMax1 = CoxaMax;
        private const int CRfFemurMin1 = FemurMin;
        private const int CRfFemurMax1 = FemurMax;
        private const int CRfTibiaMin1 = TibiaMin;
        private const int CRfTibiaMax1 = TibiaMax;
        private const int CLrCoxaMin1 = CoxaMin; //Mechanical limits of the Left Rear Leg
        private const int CLrCoxaMax1 = CoxaMax;
        private const int CLrFemurMin1 = FemurMin;
        private const int CLrFemurMax1 = FemurMax;
        private const int CLrTibiaMin1 = TibiaMin;
        private const int CLrTibiaMax1 = TibiaMax;
        private const int CLmCoxaMin1 = CoxaMin; //Mechanical limits of the Left Middle Leg
        private const int CLmCoxaMax1 = CoxaMax;
        private const int CLmFemurMin1 = FemurMin;
        private const int CLmFemurMax1 = FemurMax;
        private const int CLmTibiaMin1 = TibiaMin;
        private const int CLmTibiaMax1 = TibiaMax;
        private const int CLfCoxaMin1 = CoxaMin; //Mechanical limits of the Left Front Leg
        private const int CLfCoxaMax1 = CoxaMax;
        private const int CLfFemurMin1 = FemurMin;
        private const int CLfFemurMax1 = FemurMax;
        private const int CLfTibiaMin1 = TibiaMin;
        private const int CLfTibiaMax1 = TibiaMax;
        private const double CRrCoxaAngle1 = -450;
        private const double CRmCoxaAngle1 = 0;
        private const double CRfCoxaAngle1 = 450;
        private const double CLrCoxaAngle1 = -450;
        private const double CLmCoxaAngle1 = 0;
        private const double CLfCoxaAngle1 = 450;
        private const double CRfOffsetX = -70;
        private const double CRfOffsetZ = -120;
        private const double CLfOffsetZ = -120;
        private const double CLfOffsetX = 70;
        private const double CRrOffsetZ = 120;
        private const double CRrOffsetX = -70;
        private const double CLrOffsetZ = 120;
        private const double CLrOffsetX = 70;
        private const double CRmOffsetX = -140;
        private const double CRmOffsetZ = 0;
        private const double CLmOffsetX = 140;
        private const double CLmOffsetZ = 0;
        private const double CoxaLength = 30; //30
        private const double FemurLength = 60; //60
        private const double TibiaLength = 125; //70
        private const double CRrCoxaLength = CoxaLength;
        private const double CRmCoxaLength = CoxaLength;
        private const double CRfCoxaLength = CoxaLength;
        private const double CLrCoxaLength = CoxaLength;
        private const double CLmCoxaLength = CoxaLength;
        private const double CLfCoxaLength = CoxaLength;
        private const double CRrFemurLength = FemurLength;
        private const double CRmFemurLength = FemurLength;
        private const double CRfFemurLength = FemurLength;
        private const double CLrFemurLength = FemurLength;
        private const double CLmFemurLength = FemurLength;
        private const double CLfFemurLength = FemurLength;
        private const double CLfTibiaLength = TibiaLength;
        private const double CLmTibiaLength = TibiaLength;
        private const double CLrTibiaLength = TibiaLength;
        private const double CRfTibiaLength = TibiaLength;
        private const double CRmTibiaLength = TibiaLength;
        private const double CRrTibiaLength = TibiaLength;
        private const double CHexInitXz = 105;
        private const double CHexInitXzCos45 = 74; // COS(45) = .7071
        private const double CHexInitXzSin45 = 74; // sin(45) = .7071
        private const double CHexInitY = 26;
        private const double CRfInitPosX = CHexInitXzCos45;
        private const double CRfInitPosY = CHexInitY;
        private const double CRfInitPosZ = -CHexInitXzSin45;
        private const double CLrInitPosX = CHexInitXzCos45;
        private const double CLrInitPosY = CHexInitY;
        private const double CLrInitPosZ = CHexInitXzCos45;
        private const double CLmInitPosX = CHexInitXz;
        private const double CLmInitPosY = CHexInitY;
        private const double CLmInitPosZ = 0;
        private const double CLfInitPosX = CHexInitXzCos45;
        private const double CLfInitPosY = CHexInitY;
        private const double CLfInitPosZ = -CHexInitXzSin45;
        private const double CRmInitPosX = CHexInitXz;
        private const double CRmInitPosY = CHexInitY;
        private const double CRmInitPosZ = 0;
        private const double CRrInitPosX = CHexInitXzCos45;
        private const double CRrInitPosY = CHexInitY;
        private const double CRrInitPosZ = CHexInitXzSin45;

        private readonly double[] _cCoxaAngle1 =
        {
            CRrCoxaAngle1, CRmCoxaAngle1, CRfCoxaAngle1, CLrCoxaAngle1, CLmCoxaAngle1, CLfCoxaAngle1
        };

        private readonly double[] _cCoxaLength =
        {
            CRrCoxaLength, CRmCoxaLength, CRfCoxaLength, CLrCoxaLength, CLmCoxaLength, CLfCoxaLength
        };

        private readonly int[] _cCoxaMax1 =
        {
            CRrCoxaMax1, CRmCoxaMax1, CRfCoxaMax1, CLrCoxaMax1, CLmCoxaMax1, CLfCoxaMax1
        };

        private readonly int[] _cCoxaMin1 =
        {
            CRrCoxaMin1, CRmCoxaMin1, CRfCoxaMin1, CLrCoxaMin1, CLmCoxaMin1, CLfCoxaMin1
        };

        private readonly short[] _cFemurHornOffset1 =
        {
            CRrFemurHornOffset1, CRmFemurHornOffset1, CRfFemurHornOffset1, CLrFemurHornOffset1, CLmFemurHornOffset1, CLfFemurHornOffset1
        };

        private readonly double[] _cFemurLength =
        {
            CRrFemurLength, CRmFemurLength, CRfFemurLength, CLrFemurLength, CLmFemurLength, CLfFemurLength
        };

        private readonly int[] _cFemurMax1 =
        {
            CRrFemurMax1, CRmFemurMax1, CRfFemurMax1, CLrFemurMax1, CLmFemurMax1, CLfFemurMax1
        };

        private readonly int[] _cFemurMin1 =
        {
            CRrFemurMin1, CRmFemurMin1, CRfFemurMin1, CLrFemurMin1, CLmFemurMin1, CLfFemurMin1
        };

        private readonly double[] _cInitPosX =
        {
            CRrInitPosX, CRmInitPosX, CRfInitPosX, CLrInitPosX, CLmInitPosX, CLfInitPosX
        };

        private readonly double[] _cInitPosY =
        {
            CRrInitPosY, CRmInitPosY, CRfInitPosY, CLrInitPosY, CLmInitPosY, CLfInitPosY
        };

        private readonly double[] _cInitPosZ =
        {
            CRrInitPosZ, CRmInitPosZ, CRfInitPosZ, CLrInitPosZ, CLmInitPosZ, CLfInitPosZ
        };

        private readonly double[] _cOffsetX = {CRrOffsetX, CRmOffsetX, CRfOffsetX, CLrOffsetX, CLmOffsetX, CLfOffsetX};
        private readonly double[] _cOffsetZ = {CRrOffsetZ, CRmOffsetZ, CRfOffsetZ, CLrOffsetZ, CLmOffsetZ, CLfOffsetZ};
        private readonly double[] _coxaAngle1 = new double[6];

        private readonly double[] _cTibiaLength =
        {
            CRrTibiaLength, CRmTibiaLength, CRfTibiaLength, CLrTibiaLength, CLmTibiaLength, CLfTibiaLength
        };

        private readonly int[] _cTibiaMax1 =
        {
            CRrTibiaMax1, CRmTibiaMax1, CRfTibiaMax1, CLrTibiaMax1, CLmTibiaMax1, CLfTibiaMax1
        };

        private readonly int[] _cTibiaMin1 =
        {
            CRrTibiaMin1, CRmTibiaMin1, CRfTibiaMin1, CLrTibiaMin1, CLmTibiaMin1, CLfTibiaMin1
        };

        private readonly double[] _femurAngle1 = new double[6]; //Actual Angle of the vertical hip, decimals = 1
        private readonly byte[] _gaitLegNr = new byte[6]; //Init position of the leg
        private readonly double[] _gaitPosX = new double[6]; //Array containing Relative X position corresponding to the Gait
        private readonly double[] _gaitPosY = new double[6]; //Array containing Relative Y position corresponding to the Gait
        private readonly double[] _gaitPosZ = new double[6]; //Array containing Relative Z position corresponding to the Gait
        private readonly double[] _gaitRotY = new double[6]; //Array containing Relative Y rotation corresponding to the Gait
        //ArcCosinus Table
        //Table build in to 3 part to get higher accuracy near cos = 1. 
        //The biggest error is near cos = 1 and has a biggest value of 3*0.012098rad = 0.521 deg.
        //-    Cos 0 to 0.9 is done by steps of 0.0079 rad. [1/127]
        //-    Cos 0.9 to 0.99 is done by steps of 0.0008 rad [0.1/127]
        //-    Cos 0.99 to 1 is done by step of 0.0002 rad [0.01/64]
        //Since the tables are overlapping the full range of 127+127+64 is not necessary. Total bytes: 277
        private readonly double[] _getACos =
        {
            255, 254, 252, 251, 250, 249, 247, 246, 245, 243, 242, 241, 240, 238, 237, 236, 234, 233, 232, 231, 229, 228, 227, 225,
            224, 223, 221, 220, 219, 217, 216, 215, 214, 212, 211, 210, 208, 207, 206, 204, 203, 201, 200, 199, 197, 196, 195, 193,
            192, 190, 189, 188, 186, 185, 183, 182, 181, 179, 178, 176, 175, 173, 172, 170, 169, 167, 166, 164, 163, 161, 160, 158,
            157, 155, 154, 152, 150, 149, 147, 146, 144, 142, 141, 139, 137, 135, 134, 132, 130, 128, 127, 125, 123, 121, 119, 117,
            115, 113, 111, 109, 107, 105, 103, 101, 98, 96, 94, 92, 89, 87, 84, 81, 79, 76, 73, 73, 73, 72, 72, 72, 71, 71, 71, 70, 70,
            70, 70, 69, 69, 69, 68, 68, 68, 67, 67, 67, 66, 66, 66, 65, 65, 65, 64, 64, 64, 63, 63, 63, 62, 62, 62, 61, 61, 61, 60, 60, 59,
            59, 59, 58, 58, 58, 57, 57, 57, 56, 56, 55, 55, 55, 54, 54, 53, 53, 53, 52, 52, 51, 51, 51, 50, 50, 49, 49, 48, 48, 47, 47, 47,
            46, 46, 45, 45, 44, 44, 43, 43, 42, 42, 41, 41, 40, 40, 39, 39, 38, 37, 37, 36, 36, 35, 34, 34, 33, 33, 32, 31, 31, 30, 29, 28,
            28, 27, 26, 25, 24, 23, 23, 23, 23, 22, 22, 22, 22, 21, 21, 21, 21, 20, 20, 20, 19, 19, 19, 19, 18, 18, 18, 17, 17, 17, 17, 16,
            16, 16, 15, 15, 15, 14, 14, 13, 13, 13, 12, 12, 11, 11, 10, 10, 9, 9, 8, 7, 6, 6, 5, 3, 0
        };

        private readonly double[] _legPosX = new double[6]; //Actual X Posion of the Leg should be length of 6
        private readonly double[] _legPosY = new double[6]; //Actual Y Posion of the Leg
        private readonly double[] _legPosZ = new double[6]; //Actual Z Posion of the Leg
        private readonly double[] _tibiaAngle1 = new double[6]; //Actual Angle of the knee, decimals = 1
        private Action<string> _action;
        private int _gaitStep;
        private int _halfLiftHeigth; //If TRUE the outer positions of the ligted legs will be half height    
        private int _ikSolution; //Output true if the solution is possible
        private int _ikSolutionError; //Output true if the solution is NOT possible
        private int _ikSolutionWarning; //Output true if the solution is NEARLY possible
        private int _lastLeg; //TRUE when the current leg is the last leg of the sequence
        private byte _liftDivFactor; //Normaly: 2, when NrLiftedPos=5: 4
        private long _nrLiftedPos; //Number of positions that a single leg is lifted [1-3]
        private byte _stepsInGait; //Number of steps in gait
        private double _tlDivFactor; //Number of steps that a leg is on the floor while walking
        private bool _travelRequest; //Temp to check if the gait is in motion
        public long BalanceDivFactor = 6; //Number of legs
        public double BodyPosX; //Global Input for the position of the body
        public double BodyPosY;
        public double BodyPosZ;
        public double BodyRotOffsetX; //Input X offset value to adjust centerpoint of rotation
        public double BodyRotOffsetY; //Input Y offset value to adjust centerpoint of rotation
        public double BodyRotOffsetZ; //Input Z offset value to adjust centerpoint of rotation
        public double BodyRotX1; //Global Input pitch of the body
        public double BodyRotY1; //Global Input rotation of the body
        public double BodyRotZ1; //Global Input roll of the body
        public int GaitType;
        public double LegLiftHeight; //Current Travel height
        public int NomGaitSpeed = 120; //Nominal speed of the gait
        public bool Running;
        public double TravelLengthX; //Current Travel length X
        public double TravelLengthZ; //Current Travel length Z
        public double TravelRotationY; //Current Travel Rotation Y
        private double YawLevel;
        private double PitchLevel;
        private double RollLevel;

        public void Setup()
        {
            //Tars Init Positions
            for (var legIndex = 0; legIndex <= 5; legIndex++)
            {
                _legPosX[legIndex] = (_cInitPosX[legIndex]); //Set start positions for each leg
                _legPosY[legIndex] = (_cInitPosY[legIndex]);
                _legPosZ[legIndex] = (_cInitPosZ[legIndex]);
            }

            GaitType = 0;
            _gaitStep = 0;
        }

        public static double Map(double x, double inMin, double inMax, double outMin, double outMax)
        {
            var r = (x - inMin) * (outMax - outMin) / (inMax - inMin) + outMin;
            return r;
        }

        /// <summary>
        /// This does not work yet
        /// </summary>
        void Balance()
        {
            if (BalanceMode)
            {
                if (_sensor.Roll > RollLevel + .8)
                {
                    BodyRotZ1 = -(Map(_sensor.Roll, 0, 8, 0, 18));
                }
                else if (_sensor.Roll < -(RollLevel - .8))
                {
                    BodyRotZ1 = -(Map(_sensor.Roll, 0, 8, 0, 18));
                }
                else
                    BodyRotZ1 = 0;

                if (_sensor.Pitch > PitchLevel + .8)
                {
                    BodyRotX1 = (Map(_sensor.Pitch, 0, 8, 0, 18));
                }
                else if (_sensor.Pitch < -(PitchLevel - .8))
                {
                    BodyRotX1 = -(Map(_sensor.Pitch, 0, 8, 0, 18));
                }
                else
                    BodyRotX1 = 0;
            }
        }

        /// <summary>
        /// Does not smoothly transition gaits yet. May stop mid step
        /// </summary>
        /// <param name="callBack"></param>
        public void Run(Action<string> callBack)
        {
            LegLiftHeight = 40;
            BodyPosY = 80;
            GaitType = 0;
            GaitSelect();

            RollLevel = _sensor.Roll;
            PitchLevel = _sensor.Pitch;

            var t = new Thread(() =>
            {
                _action = callBack;

                while (Running)
                {

                    Balance();

                    //Debug.WriteLine("TravelLengthZ " + TravelLengthZ);
                    //Debug.WriteLine("TravelLengthX " + TravelLengthX);
                    //Debug.WriteLine("TravelRotationY " + TravelRotationY);
                    //Gait
                    GaitSeq();

                    //Reset IKsolution indicators 
                    _ikSolution = 0;
                    _ikSolutionWarning = 0;
                    _ikSolutionError = 0;

                    double bodyFkPosX;
                    double bodyFkPosY;
                    double bodyFkPosZ;

                    //if (_sensor.WarningCenter && TravelLengthZ < 0)
                    //    TravelLengthZ = 0;

                    //Do IK for all Right legs
                    for (var legIndex = 0; legIndex <= 2; legIndex++)
                    {
                        BodyFk(-_legPosX[legIndex] + BodyPosX + _gaitPosX[legIndex],
                            _legPosZ[legIndex] + BodyPosZ + _gaitPosZ[legIndex],
                            _legPosY[legIndex] + BodyPosY + _gaitPosY[legIndex],
                            _gaitRotY[legIndex], legIndex,
                            out bodyFkPosX, out bodyFkPosZ, out bodyFkPosY);

                        LegIk(_legPosX[legIndex] - BodyPosX + bodyFkPosX - (_gaitPosX[legIndex]),
                            _legPosY[legIndex] + BodyPosY - bodyFkPosY + _gaitPosY[legIndex],
                            _legPosZ[legIndex] + BodyPosZ - bodyFkPosZ + _gaitPosZ[legIndex],
                            legIndex);
                    }

                    //Do IK for all Left legs  
                    for (var legIndex = 3; legIndex <= 5; legIndex++)
                    {
                        BodyFk(_legPosX[legIndex] - BodyPosX + _gaitPosX[legIndex],
                            _legPosZ[legIndex] + BodyPosZ + _gaitPosZ[legIndex],
                            _legPosY[legIndex] + BodyPosY + _gaitPosY[legIndex],
                            _gaitRotY[legIndex], legIndex,
                            out bodyFkPosX, out bodyFkPosZ, out bodyFkPosY);

                        LegIk(_legPosX[legIndex] + BodyPosX - bodyFkPosX + _gaitPosX[legIndex],
                            _legPosY[legIndex] + BodyPosY - bodyFkPosY + _gaitPosY[legIndex],
                            _legPosZ[legIndex] + BodyPosZ - bodyFkPosZ + _gaitPosZ[legIndex],
                            legIndex);
                    }

                    //Debug.WriteLine("IKSolution " + _ikSolution);
                    //Debug.WriteLine("IKSolutionWarning " + _ikSolutionWarning);
                    //Debug.WriteLine("IKSolutionError " + _ikSolutionError);

                    if (_ikSolutionError == 1 || _ikSolution == 0)
                        continue;

                    CheckAngles();

                    Thread.Sleep(NomGaitSpeed - 15);

                    UpdateServoDriver();
                }
            }) {IsBackground = true};
            t.Start();
        }

        public bool BalanceMode { get; set; }

        public void GaitSelect()
        {
            //Gait selector
            switch (GaitType)
            {
                case 0:
                    //Ripple Gait 12 steps
                    _gaitLegNr[CLr] = 1;
                    _gaitLegNr[CRf] = 3;
                    _gaitLegNr[CLm] = 5;
                    _gaitLegNr[CRr] = 7;
                    _gaitLegNr[CLf] = 9;
                    _gaitLegNr[CRm] = 11;

                    _nrLiftedPos = 3;
                    _halfLiftHeigth = 3;
                    _tlDivFactor = 8;
                    _stepsInGait = 12;
                    //NomGaitSpeed = 110;
                    break;
                case 1:
                    //Tripod 8 steps
                    _gaitLegNr[CLr] = 5;
                    _gaitLegNr[CRf] = 1;
                    _gaitLegNr[CLm] = 1;
                    _gaitLegNr[CRr] = 1;
                    _gaitLegNr[CLf] = 5;
                    _gaitLegNr[CRm] = 5;

                    _nrLiftedPos = 3;
                    _halfLiftHeigth = 3;
                    _tlDivFactor = 4;
                    _stepsInGait = 8;
                    //NomGaitSpeed = 80;
                    break;
                case 2:
                    //Triple Tripod 12 step
                    _gaitLegNr[CRf] = 3;
                    _gaitLegNr[CLm] = 4;
                    _gaitLegNr[CRr] = 5;
                    _gaitLegNr[CLf] = 9;
                    _gaitLegNr[CRm] = 10;
                    _gaitLegNr[CLr] = 11;

                    _nrLiftedPos = 3;
                    _halfLiftHeigth = 3;
                    _tlDivFactor = 8;
                    _stepsInGait = 12;
                    //NomGaitSpeed = 100;
                    break;
                case 3:
                    // Triple Tripod 16 steps, use 5 lifted positions
                    _gaitLegNr[CRf] = 4;
                    _gaitLegNr[CLm] = 5;
                    _gaitLegNr[CRr] = 6;
                    _gaitLegNr[CLf] = 12;
                    _gaitLegNr[CRm] = 13;
                    _gaitLegNr[CLr] = 14;

                    _nrLiftedPos = 5;
                    _halfLiftHeigth = 1;
                    _tlDivFactor = 10;
                    _stepsInGait = 16;
                    //NomGaitSpeed = 100;
                    break;
                case 4:
                    //Wave 24 steps
                    _gaitLegNr[CLr] = 1;
                    _gaitLegNr[CRf] = 21;
                    _gaitLegNr[CLm] = 5;

                    _gaitLegNr[CRr] = 13;
                    _gaitLegNr[CLf] = 9;
                    _gaitLegNr[CRm] = 17;

                    _nrLiftedPos = 3;
                    _halfLiftHeigth = 3;
                    _tlDivFactor = 20;
                    _stepsInGait = 24;
                    //NomGaitSpeed = 110;
                    break;
            }
        }

        private void GaitSeq()
        {
            //Check if the Gait is in motion
            _travelRequest = ((Math.Abs(TravelLengthX) > CTravelDeadZone) || (Math.Abs(TravelLengthZ) > CTravelDeadZone) || (Math.Abs(TravelRotationY) > CTravelDeadZone));

            if (_nrLiftedPos == 5)
                _liftDivFactor = 4;
            else
                _liftDivFactor = 2;

            //Calculate Gait sequence
            _lastLeg = 0;
            for (var legIndex = 0; legIndex <= 5; legIndex++)
            {
                // for all legs
                if (legIndex == 5) // last leg
                    _lastLeg = 1;

                Gait(legIndex);
            } // next leg
        }

        private void Gait(int gaitCurrentLegNr)
        {
            //Clear values under the cTravelDeadZone
            if (!_travelRequest)
            {
                TravelLengthX = 0;
                TravelLengthZ = 0;
                TravelRotationY = 0;
            }
            //Leg middle up position
            //Gait in motion														  									
            //Gait NOT in motion, return to home position
            if ((_travelRequest && (_nrLiftedPos == 1 || _nrLiftedPos == 3 || _nrLiftedPos == 5) && _gaitStep == _gaitLegNr[gaitCurrentLegNr]) || (!_travelRequest && _gaitStep == _gaitLegNr[gaitCurrentLegNr] && ((Math.Abs(_gaitPosX[gaitCurrentLegNr]) > 2) || (Math.Abs(_gaitPosZ[gaitCurrentLegNr]) > 2) || (Math.Abs(_gaitRotY[gaitCurrentLegNr]) > 2))))
            {
                //Up
                _gaitPosX[gaitCurrentLegNr] = 0;
                _gaitPosY[gaitCurrentLegNr] = -LegLiftHeight;
                _gaitPosZ[gaitCurrentLegNr] = 0;
                _gaitRotY[gaitCurrentLegNr] = 0;
            }
            //Optional Half height Rear (2, 3, 5 lifted positions)
            else if (((_nrLiftedPos == 2 && _gaitStep == _gaitLegNr[gaitCurrentLegNr]) || (_nrLiftedPos >= 3 && (_gaitStep == _gaitLegNr[gaitCurrentLegNr] - 1 || _gaitStep == _gaitLegNr[gaitCurrentLegNr] + (_stepsInGait - 1)))) && _travelRequest)
            {
                _gaitPosX[gaitCurrentLegNr] = -TravelLengthX/_liftDivFactor;
                _gaitPosY[gaitCurrentLegNr] = -3*LegLiftHeight/(3 + _halfLiftHeigth);
                //Easier to shift between div factor: /1 (3/3), /2 (3/6) and 3/4
                _gaitPosZ[gaitCurrentLegNr] = -TravelLengthZ/_liftDivFactor;
                _gaitRotY[gaitCurrentLegNr] = -TravelRotationY/_liftDivFactor;
            }

            // Optional Half height front (2, 3, 5 lifted positions)
            else if ((_nrLiftedPos >= 2) && (_gaitStep == _gaitLegNr[gaitCurrentLegNr] + 1 || _gaitStep == _gaitLegNr[gaitCurrentLegNr] - (_stepsInGait - 1)) && _travelRequest)
            {
                _gaitPosX[gaitCurrentLegNr] = TravelLengthX/_liftDivFactor;
                _gaitPosY[gaitCurrentLegNr] = -3*LegLiftHeight/(3 + _halfLiftHeigth);
                // Easier to shift between div factor: /1 (3/3), /2 (3/6) and 3/4
                _gaitPosZ[gaitCurrentLegNr] = TravelLengthZ/_liftDivFactor;
                _gaitRotY[gaitCurrentLegNr] = TravelRotationY/_liftDivFactor;
            }

            //Optional Half heigth Rear 5 LiftedPos (5 lifted positions)
            else if (((_nrLiftedPos == 5 && (_gaitStep == _gaitLegNr[gaitCurrentLegNr] - 2))) && _travelRequest)
            {
                _gaitPosX[gaitCurrentLegNr] = -TravelLengthX/2;
                _gaitPosY[gaitCurrentLegNr] = -LegLiftHeight/2;
                _gaitPosZ[gaitCurrentLegNr] = -TravelLengthZ/2;
                _gaitRotY[gaitCurrentLegNr] = -TravelRotationY/2;
            }

            //Optional Half heigth Front 5 LiftedPos (5 lifted positions)
            else if ((_nrLiftedPos == 5) && (_gaitStep == _gaitLegNr[gaitCurrentLegNr] + 2 || _gaitStep == _gaitLegNr[gaitCurrentLegNr] - (_stepsInGait - 2)) && _travelRequest)
            {
                _gaitPosX[gaitCurrentLegNr] = TravelLengthX/2;
                _gaitPosY[gaitCurrentLegNr] = -LegLiftHeight/2;
                _gaitPosZ[gaitCurrentLegNr] = TravelLengthZ/2;
                _gaitRotY[gaitCurrentLegNr] = TravelRotationY/2;
            }

            //Leg front down position
            else if ((_gaitStep == _gaitLegNr[gaitCurrentLegNr] + _nrLiftedPos || _gaitStep == _gaitLegNr[gaitCurrentLegNr] - (_stepsInGait - _nrLiftedPos)) && _gaitPosY[gaitCurrentLegNr] < 0)
            {
                _gaitPosX[gaitCurrentLegNr] = TravelLengthX/2;
                _gaitPosZ[gaitCurrentLegNr] = TravelLengthZ/2;
                _gaitRotY[gaitCurrentLegNr] = TravelRotationY/2;
                _gaitPosY[gaitCurrentLegNr] = 0;
                //Only move leg down at once if terrain adaption is turned off
            }

            //Move body forward      
            else
            {
                _gaitPosX[gaitCurrentLegNr] = _gaitPosX[gaitCurrentLegNr] - (TravelLengthX/_tlDivFactor);
                _gaitPosY[gaitCurrentLegNr] = 0;
                _gaitPosZ[gaitCurrentLegNr] = _gaitPosZ[gaitCurrentLegNr] - (TravelLengthZ/_tlDivFactor);
                _gaitRotY[gaitCurrentLegNr] = _gaitRotY[gaitCurrentLegNr] - (TravelRotationY/_tlDivFactor);
            }

            //Advance to the next step
            if (_lastLeg == 1)
            {
                //The last leg in this step
                _gaitStep = _gaitStep + 1;
                if (_gaitStep > _stepsInGait)
                    _gaitStep = 1;
            }
        }

        //[GETSINCOS] Get the sinus and cosinus from the angle +/- multiple circles
        //AngleDeg1     - Input Angle in degrees
        //sin4        - Output Sinus of AngleDeg
        //cos4          - Output Cosinus of AngleDeg
        private static void GetSinCos(double angleDeg1, out double sin, out double cos)
        {
            var angle = Math.PI*angleDeg1/180.0;

            sin = Math.Round(Math.Sin(angle)*10000);
            cos = Math.Round(Math.Cos(angle)*10000);
        }

        //(GETARCCOS) Get the sinus and cosinus from the angle +/- multiple circles
        //cos4        - Input Cosinus
        //AngleRad4     - Output Angle in AngleRad4
        private double GetArcCos(double cos4)
        {
            bool negativeValue; //If the the value is Negative
            double angleRad4 = 0;

            if (cos4 < 0)
            {
                cos4 = -cos4;
                negativeValue = true;
            }
            else
                negativeValue = false;

            //Limit cos4 to his maximal value
            cos4 = Math.Min(cos4, C4Dec);

            if ((cos4 >= 0) && (cos4 < 9000))
            {
                angleRad4 = (_getACos[Convert.ToInt32(cos4/79)]);
                angleRad4 = (angleRad4*616)/C1Dec; //616=acos resolution (pi/2/255) ;
            }
            else if ((cos4 >= 9000) && (cos4 < 9900))
            {
                angleRad4 = (_getACos[Convert.ToInt32((cos4 - 9000)/8 + 114)]);
                angleRad4 = angleRad4*616/C1Dec; //616=acos resolution (pi/2/255) 
            }
            else if ((cos4 >= 9900) && (cos4 <= 10000))
            {
                angleRad4 = (_getACos[Convert.ToInt32((cos4 - 9900)/2 + 227)]);
                angleRad4 = angleRad4*616/C1Dec; //616=acos resolution (pi/2/255) 
            }

            //Add negative sign
            if (negativeValue)
                angleRad4 = 31416 - angleRad4;

            return angleRad4;
        }

        //(GETATAN2) Simplyfied ArcTan2 function based on fixed point ArcCos
        //ArcTanX         - Input X
        //ArcTanY         - Input Y
        //ArcTan4          - Output ARCTAN2(X/Y)
        //XYhyp2            - Output presenting Hypotenuse of X and Y
        private double GetATan2(double atanX, double atanY, out double xyhyp2)
        {
            double atan4;

            xyhyp2 = Math.Sqrt((atanX*atanX*C4Dec) + (atanY*atanY*C4Dec)); //FIXED

            var angleRad4 = GetArcCos((atanX*C6Dec)/xyhyp2);

            if (atanY < 0) // removed overhead... Atan4 = AngleRad4 * (AtanY/abs(AtanY));  
                atan4 = -angleRad4;
            else
                atan4 = angleRad4;

            return atan4;
        }

        //(BODY INVERSE KINEMATICS) 
        //BodyRotX         - Global Input pitch of the body 
        //BodyRotY         - Global Input rotation of the body 
        //BodyRotZ         - Global Input roll of the body 
        //RotationY         - Input Rotation for the gait 
        //PosX            - Input position of the feet X 
        //PosZ            - Input position of the feet Z 
        //SinB                  - Sin buffer for BodyRotX
        //CosB               - Cos buffer for BodyRotX 
        //SinG                  - Sin buffer for BodyRotZ
        //CosG               - Cos buffer for BodyRotZ
        //BodyFKPosX         - Output Position X of feet with Rotation 
        //BodyFKPosY         - Output Position Y of feet with Rotation 
        //BodyFKPosZ         - Output Position Z of feet with Rotation
        private void BodyFk(double posX, double posZ, double posY, double rotationY, int bodyIkLeg, out double bodyFkPosX, out double bodyFkPosZ, out double bodyFkPosY)
        {
            double sinA4; //Sin buffer for BodyRotX calculations
            double cosA4; //Cos buffer for BodyRotX calculations
            double sinB4; //Sin buffer for BodyRotX calculations
            double cosB4; //Cos buffer for BodyRotX calculations
            double sinG4; //Sin buffer for BodyRotZ calculations
            double cosG4; //Cos buffer for BodyRotZ calculations

            //Calculating totals from center of the body to the feet 
            var cprX = (_cOffsetX[bodyIkLeg]) + posX + BodyRotOffsetX;
            var cprY = posY + BodyRotOffsetY;
            var cprZ = (_cOffsetZ[bodyIkLeg]) + posZ + BodyRotOffsetZ;

            //Successive global rotation matrix: 
            //Math shorts for rotation: Alfa [A] = Xrotate, Beta [B] = Zrotate, Gamma [G] = Yrotate 
            //Sinus Alfa = SinA, cosinus Alfa = cosA. and so on... 

            //First calculate sinus and cosinus for each rotation: 
            GetSinCos(BodyRotX1, out sinG4, out cosG4);

            GetSinCos(BodyRotZ1, out sinB4, out cosB4);

            GetSinCos(BodyRotY1 + (rotationY*C1Dec), out sinA4, out cosA4);

            //Calculation of rotation matrix: 
            bodyFkPosX = (cprX*C2Dec - ((cprX*C2Dec*cosA4/C4Dec*cosB4/C4Dec) - (cprZ*C2Dec*cosB4/C4Dec*sinA4/C4Dec) + (cprY*C2Dec*sinB4/C4Dec)))/C2Dec;
            bodyFkPosZ = (cprZ*C2Dec - ((cprX*C2Dec*cosG4/C4Dec*sinA4/C4Dec) + (cprX*C2Dec*cosA4/C4Dec*sinB4/C4Dec*sinG4/C4Dec) + (cprZ*C2Dec*cosA4/C4Dec*cosG4/C4Dec) - (cprZ*C2Dec*sinA4/C4Dec*sinB4/C4Dec*sinG4/C4Dec) - (cprY*C2Dec*cosB4/C4Dec*sinG4/C4Dec)))/C2Dec;
            bodyFkPosY = (cprY*C2Dec - ((cprX*C2Dec*sinA4/C4Dec*sinG4/C4Dec) - (cprX*C2Dec*cosA4/C4Dec*cosG4/C4Dec*sinB4/C4Dec) + (cprZ*C2Dec*cosA4/C4Dec*sinG4/C4Dec) + (cprZ*C2Dec*cosG4/C4Dec*sinA4/C4Dec*sinB4/C4Dec) + (cprY*C2Dec*cosB4/C4Dec*cosG4/C4Dec)))/C2Dec;
        }

        //[LEG INVERSE KINEMATICS] Calculates the angles of the coxa, femur and tibia for the given position of the feet
        //IKFeetPosX            - Input position of the Feet X
        //IKFeetPosY            - Input position of the Feet Y
        //IKFeetPosZ            - Input Position of the Feet Z
        //IKSolution            - Output true if the solution is possible
        //IKSolutionWarning     - Output true if the solution is NEARLY possible
        //IKSolutionError    - Output true if the solution is NOT possible
        //FemurAngle1           - Output Angle of Femur in degrees
        //TibiaAngle1           - Output Angle of Tibia in degrees
        //CoxaAngle1            - Output Angle of Coxa in degrees
        private void LegIk(double ikFeetPosX, double ikFeetPosY, double ikFeetPosZ, int legIkLegNr)
        {
            double xyhyp2;

            //Calculate IKCoxaAngle and IKFeetPosXZ
            var getatan = GetATan2(ikFeetPosX, ikFeetPosZ, out xyhyp2);
            _coxaAngle1[legIkLegNr] = ((getatan*180)/3141) + (_cCoxaAngle1[legIkLegNr]);

            var ikFeetPosXz = xyhyp2/C2Dec;
            var ika14 = GetATan2(ikFeetPosY, ikFeetPosXz - (_cCoxaLength[legIkLegNr]), out xyhyp2);
            var iksw2 = xyhyp2;
            var temp1 = ((((long) (_cFemurLength[legIkLegNr])*(_cFemurLength[legIkLegNr])) - ((long) (_cTibiaLength[legIkLegNr])*(_cTibiaLength[legIkLegNr])))*C4Dec + (iksw2*iksw2));
            var temp2 = 2*(_cFemurLength[legIkLegNr])*C2Dec*iksw2;
            var ika24 = GetArcCos(temp1/(temp2/C4Dec));

            _femurAngle1[legIkLegNr] = -(ika14 + ika24)*180/3141 + 900 + _cFemurHornOffset1[legIkLegNr];

            temp1 = ((((long) (_cFemurLength[legIkLegNr])*(_cFemurLength[legIkLegNr])) + ((long) (_cTibiaLength[legIkLegNr])*(_cTibiaLength[legIkLegNr])))*C4Dec - (iksw2*iksw2));
            temp2 = (2*(_cFemurLength[legIkLegNr])*(_cTibiaLength[legIkLegNr]));

            _tibiaAngle1[legIkLegNr] = -(900 - GetArcCos(temp1/temp2)*180/3141);

            if (iksw2 < ((_cFemurLength[legIkLegNr]) + (_cTibiaLength[legIkLegNr]) - 30)*C2Dec)
                _ikSolution = 1;
            else
            {
                if (iksw2 < ((_cFemurLength[legIkLegNr]) + (_cTibiaLength[legIkLegNr]))*C2Dec)
                    _ikSolutionWarning = 1;
                else
                    _ikSolutionError = 1;
            }
        }

        //[CHECK ANGLES] Checks the mechanical limits of the servos
        private void CheckAngles()
        {
            for (var legIndex = 0; legIndex <= 5; legIndex++)
            {
                _coxaAngle1[legIndex] = Math.Min(Math.Max(_coxaAngle1[legIndex], (_cCoxaMin1[legIndex])), (_cCoxaMax1[legIndex]));
                _femurAngle1[legIndex] = Math.Min(Math.Max(_femurAngle1[legIndex], (_cFemurMin1[legIndex])), (_cFemurMax1[legIndex]));
                _tibiaAngle1[legIndex] = Math.Min(Math.Max(_tibiaAngle1[legIndex], (_cTibiaMin1[legIndex])), (_cTibiaMax1[legIndex]));
            }
        }

        // A PWM/deg factor of 10,09 give cPwmDiv = 991 and cPFConst = 592
        // For a modified 5645 (to 180 deg travel): cPwmDiv = 1500 and cPFConst = 900.
        private void UpdateServoDriver()
        {
            var sscString = string.Empty;
            var debugString = string.Empty;

            for (var legIndex = 0; legIndex <= 5; legIndex++)
            {
                if (legIndex < 3)
                {
                    var wCoxaSscv = Math.Round((-_coxaAngle1[legIndex] + 900)*1000/CPwmDiv + CPfConst);
                    var wFemurSscv = Math.Round((-_femurAngle1[legIndex] + 900)*1000/CPwmDiv + CPfConst);
                    var wTibiaSscv = Math.Round((-_tibiaAngle1[legIndex] + 900)*1000/CPwmDiv + CPfConst);

                    if (legIndex == 0)
                    {
                        sscString += "#" + _control.LegOne.CoxaServo + "P" + wCoxaSscv;
                        sscString += "#" + _control.LegOne.FemurServo + "P" + wFemurSscv;
                        sscString += "#" + _control.LegOne.TibiaServo + "P" + wTibiaSscv;
                    }
                    if (legIndex == 1)
                    {
                        sscString += "#" + _control.LegTwo.CoxaServo + "P" + wCoxaSscv;
                        sscString += "#" + _control.LegTwo.FemurServo + "P" + wFemurSscv;
                        sscString += "#" + _control.LegTwo.TibiaServo + "P" + wTibiaSscv;
                    }
                    if (legIndex == 2)
                    {
                        sscString += "#" + _control.LegThree.CoxaServo + "P" + wCoxaSscv;
                        sscString += "#" + _control.LegThree.FemurServo + "P" + wFemurSscv;
                        sscString += "#" + _control.LegThree.TibiaServo + "P" + wTibiaSscv;
                    }

                    debugString += "left leg " + legIndex + Environment.NewLine;
                    debugString += "coxa SSC " + wCoxaSscv + Environment.NewLine;
                    debugString += "femur SSC " + wFemurSscv + Environment.NewLine;
                    debugString += "tibia SSC " + wTibiaSscv + Environment.NewLine;
                }
                else
                {
                    var wCoxaSscv = Math.Round((_coxaAngle1[legIndex] + 900)*1000/CPwmDiv + CPfConst);
                    var wFemurSscv = Math.Round(((_femurAngle1[legIndex] + 900)*1000/CPwmDiv + CPfConst));
                    var wTibiaSscv = Math.Round((_tibiaAngle1[legIndex] + 900)*1000/1059 + CPfConst);

                    if (legIndex == 5)
                    {
                        sscString += "#" + _control.LegFour.CoxaServo + "P" + wCoxaSscv;
                        sscString += "#" + _control.LegFour.FemurServo + "P" + wFemurSscv;
                        sscString += "#" + _control.LegFour.TibiaServo + "P" + wTibiaSscv;
                    }
                    if (legIndex == 4)
                    {
                        sscString += "#" + _control.LegFive.CoxaServo + "P" + wCoxaSscv;
                        sscString += "#" + _control.LegFive.FemurServo + "P" + wFemurSscv;
                        sscString += "#" + _control.LegFive.TibiaServo + "P" + wTibiaSscv;
                    }
                    if (legIndex == 3)
                    {
                        sscString += "#" + _control.LegSix.CoxaServo + "P" + wCoxaSscv;
                        sscString += "#" + _control.LegSix.FemurServo + "P" + wFemurSscv;
                        sscString += "#" + _control.LegSix.TibiaServo + "P" + wTibiaSscv;
                    }

                    debugString += "right leg " + legIndex + Environment.NewLine;
                    debugString += "coxa SSC " + wCoxaSscv + Environment.NewLine;
                    debugString += "femur SSC " + wFemurSscv + Environment.NewLine;
                    debugString += "tibia SSC " + wTibiaSscv + Environment.NewLine;
                }
            }

            sscString += "T" + (NomGaitSpeed);

            _control.DirectWrite(sscString);

            if (_action != null)
                _action(debugString);
        }
    }
}